from torch import FloatTensor, Tensor
from PIL.Image import Image as ImageType
from numpy import ndarray
from typing import (
    Any, 
    Dict, 
    List, 
    Optional, 
    Union, 
    Callable,
    Tuple, 
    Iterable, 
    Type, 
    )